#!/usr/bin/sed -nrf

#################################################### WARNING ####################################################
#                               Back-up any file before you open it in this editor!!                            #
#################################################################################################################

:_Problem
    # text editor
    # https://cg.esolangs.gay/62/
:_Restrictions
    # run as: FILE="file.txt" ./cg62_editor.sed (stdin used for input of commands)
    # escape character (\o033) and "EOF" keyword are reserved, can't be in the file
    # errors aren't handled, they close the editor instantly (no save!)
    # vertical scrolling isn't implemented, cursor can go out of screen (use file with few lines to test)


################################################# SETUP #########################################################

:init
	1v;1x
	1b f5
1!b apply_cmd

################################################# COMMANDS ######################################################

:apply_cmd
	/^u$/		{	x;b u		}
	/^uu$/		{	x;b uu		}
	/^d$/		{	x;b d		}
	/^dd$/		{	x;b dd		}
	/^l$/		{	x;b l		}
	/^ll$/		{	x;b ll		}
	/^r$/		{	x;b r		}
	/^rr$/		{	x;b rr		}
	/^tab$/		{	x;b tab		}
	/^enter$/	{	x;b enter	}
	/^i ./		{	x;b i		}
	/^k$/		{	x;b k		}
	/^kk$/		{	x;b kk		}
	/^kkk$/		{	x;b kkk		}
	/^clear$/	{	x;b clear	}
	/^f5$/		{	x;b f5		}
	/^s$/		{	x;b s		}
	/^quit$/	{	x;b quit	}
	/^help$/	{	x;b help	}
	s@.*@Input error: unknown command!.@p
b quit_error

:u
	/^0/b display
	s:([^\n]*\n?[^\n]*)(\o033\[91m>\o033\[0m):\2\1:
	s:^(\o033\[91m>\o033\[0m)1:1\1:
b display

:uu
	/^1/s:^1(.*)(\o033\[91m>\o033\[0m):1\2\1:
b display

:d
	/^1/s:(\o033\[91m>\o033\[0m)([^\n]*\n?):\2\1:
b display

:dd
	/^1/s:(\o033\[91m>\o033\[0m)(.*):\2\1:
b display

:l
	/^1/s:^(1.*)(.)(\o033\[91m>\o033\[0m):\1\3\2:
b display

:ll
	/^0/b display
	s:([^\n]*)(\o033\[91m>\o033\[0m):\2\1:
	s:^(\o033\[91m>\o033\[0m)1:1\1:
b display

:r
	/^1/s:(\o033\[91m>\o033\[0m)(.):\2\1:
b display

:rr
	/^1/s:(\o033\[91m>\o033\[0m)([^\n]*):\2\1:
b display

:tab
	/^1/s:\o033\[91m>\o033\[0m:\t&:
b display

:enter
	/^1/s:\o033\[91m>\o033\[0m:\n&:
b display

:i
	/^0/b display
	x
	s:^i ::
	/\o033/{
		s@.*@Input error: ESCAPE is a reserved character!@p
		b quit_error
	}
	/EOF/{
		s@.*@Input error: 'EOF' is a reserved keyword!@p
		b quit_error
	}
	G;h;x
	s:^([^\n]+)\n(1.*)(\o033\[91m>\o033\[0m):\2\1\3:
b display

:k
	/^1/s:(\o033\[91m>\o033\[0m).:\1:
b display

:kk
	/^1/s:(\o033\[91m>\o033\[0m)[^\n]*:\1:
b display

:kkk
	/^1/s:(\o033\[91m>\o033\[0m).*:\1:
b display

:clear
	/^1/s:.*:1\o033[91m>\o033[0m:
b display

:f5
	/^0/b display
	s:.*:cat "$FILE":e
	#how to test if file doesn't exist?
	/\o033/{
		s@.*@File error: ESCAPE is a reserved character!@p
		b quit_error
	}
	/EOF/{
		s@.*@File error: 'EOF' is a reserved keyword!@p
		b quit_error
	}
	s:^:1\o033[91m>\o033[0m:
b display

:s
	/^0/b display
	h;x
	s:^1::
	s:\o033\[91m>\o033\[0m::
	s:.*:cat << 'EOF' > "$FILE"\n&\nEOF:e
	x
b display

:quit
b quit_ok

:help
	/^0/{
		s::1:
		b display
	}
	s:^1:0:
b display

:get_help
	#this should have been part of :help, but then both pattern and hold space would be occupied
	#display needs one free space to call to bash and other things, so the help text is JIT generated
	s:.*:u    move cursor to the start of the line above\n:
	s:$:uu    move cursor to the start of the file\n:
	s:$:d     move cursor to the start of the line below\n:
	s:$:dd    move cursor to the end of the file\n\n:
	s:$:l     move cursor to the left\n:
	s:$:ll    move cursor to the start of the current line\n:
	s:$:r     move cursor to the right\n:
	s:$:rr    move cursor to the end of the current line\n\n:
	s:$:tab   insert tab\n:
	s:$:enter insert enter\n:
	s:$:i TXT insert TXT characters (except tab and enter)\n\n:
	s:$:k     kill character after the cursor\n:
	s:$:kk    kill all characters after the cursor till the end of the current line\n:
	s:$:kkk   kill all characters after the cursor till the end of the file\n:
	s:$:clear kill all characters from the file\n\n:
	s:$:f5    reload content from the file on disk (unsaved changes lost)\n:
	s:$:s     save content to file on disk\n:
	s:$:quit  quit, discarding all changes! (works in help mode)\n\n:
	s:$:help  toggle this help screen (works in help mode):
b continue_display

################################################# DISPLAY #######################################################

:display
	/^1/{
		h;x
		s:::
	}
	/^0/{
		x
		b get_help
		:continue_display
	}
	:number_lines_and_wrap
		#Shit! No time left, let's cheat :D
		s@.*@cat -n << 'EOF'|sed 's:^[ \t]*::;s:\t:):'|fold -$(tput cols)\n&\nEOF@e
	#clear screen
	s:^:\o033[H\o033[J:
	s@$@\n~\nEnter command (type help for a list of available commands):@p
b EOS

:get_terminal_stats
	#sed can't directly access environmental var $FILE and terminal size
	s:.*:echo "$FILE";tput cols;tput lines:e
b#?

################################################# THE END #######################################################

:quit_ok
	Q 0

:quit_error
	Q 1

:EOS

