const hex_digit = "[0-9A-Fa-f]";
const integer = "(([1-9][0-9]*|0[0-7]*|0[xX]" ++ hex_digit ++ "+)([uU](l|L|ll|LL)|(l|L|ll|LL)[uU]))";
const dexp = "([eE][+-]?[0-9]+)";
const hexp = "([pP][+-]?[0-9]+)";
const decfloat = "([0-9]+%.[0-9]*|%.[0-9]+)" ++ dexp ++ "?|[0-9]+" ++ dexp;
const hexfloat = "0[xX](" ++ hex_digit ++ "+%." ++ hex_digit ++ "*" ++ hexp ++ "?|" ++ hex_digit ++ "+" ++ hexp ++ ")";
const float = "(" ++ decfloat ++ "|" ++ hexfloat ++ ")[fF]";
// TODO fallback tainted invalid escape sequence
const hex_quad = hex_digit ++ hex_digit ++ hex_digit ++ hex_digit;
const escape = "(\\(['\"?\\abfnrtv]|[0-7][0-7]?[0-7]?|x" ++ hex_digit ++ "+|u" ++ hex_quad ++ "|U" ++ hex_quad ++ hex_quad ++ "))";

const tok = @import("tok.zig");
const RegexTokenizer = tok.RegexTokenizer;
const Token = tok.Token;

var cache: ?RegexTokenizer = null;
pub fn tokenizer() !RegexTokenizer {
    cache = cache orelse try RegexTokenizer.init(&.{
        // zig fmt: off
        .{.keyword, "auto|break|case|char|const|continue|default|do|double"
            ++ "|else|enum|extern|float|for|goto|if|inline|int|long"
            ++ "|register|restrict|return|short|signed|sizeof|static|struct|switch"
            ++ "|typedef|union|unsigned|void|volatile|while"
            ++ "|_Alignas|_Alignof|_Atomic|_Bool|_Complex|_Generic|_Imaginary|_Noreturn|_Static_assert|_Thread_local"},
        .{.identifier, "[a-zA-Z_]+[a-zA-Z0-9_]*"},
        .{.quoted, "(u8|[uUL])?\"([^\"\n\\]|" ++ escape ++ ")*\"|[uUL]?'([^'\n\\]|" ++ escape ++ ")*'"},
        .{.number, integer ++ "|" ++ float},
        .{.sym, "[(){}[%]<>.!&|+%-*/%]|%->|%+%+|%-%-|<<|>>|[!=<>]=|%^|&&|%|%||%?|:|;|%.%.%.|([*/&+%-^|]|<<|>>)?=|,|#|##"
            ++ "|<:|:>|<%%|%%>|%%:|%%:%%:"}, // darn digraphs
        .{.pp, "#include[ \t]*\"[^\n\"]*\"|<[^\n>]*>|#(pragma|define|undef|if(n?def)|else|elif|endif|line|error|warning)"},
        .{.comment, "//[^\n]*|/%*(/[^*]|[^/])*%*/"},
        .{.ignore, "[ \n\r\t]+"},
    });
    return cache.?;
}

const expectEq = @import("std").testing.expectEqual;
const Tester = struct {
    t: RegexTokenizer,
    buf: []const u8,
    i: usize = 0,
    pub fn expect(self: *@This(), kind: tok.TokenType, s: []const u8) !void {
        try expectEq(Token{.kind = kind, .len = s.len}, self.t.nextToken(self.buf[self.i..]));
        self.i += s.len;
    }
};
test "things" {
    const buf = "if (1) printf(\"blah\");";
    var t = Tester{.t = try tokenizer(), .buf = buf};
    try t.expect(.keyword, "if");
    try t.expect(.ignore, " ");
    try t.expect(.sym, "(");
    try t.expect(.number, "2");
    try t.expect(.sym, ")");
    try t.expect(.ignore, " ");
    try t.expect(.identifier, "printf");
    try t.expect(.sym, "(");
    try t.expect(.quoted, "\"blah\"");
    try t.expect(.sym, ")");
    try t.expect(.sym, ";");
}